#include <time.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <x86intrin.h>
#include <omp.h>
#include "btc.h"
//#include <simstruc.h> // needed to forward printf output to matlab

decom rank1approx_parallel(const uint64_t* restrict A, const uint64_t* restrict colsums, const dim2 dim) {
	const int integerPerColumn = dim.x / 64 + (dim.x % 64 > 0 ? 1 : 0);
	const int integerPerRow = dim.y / 64 + (dim.y % 64 > 0 ? 1 : 0);
	uint64_t bestSim = 0;
	decom bestVectors = {.A = checkedCalloc(integerPerColumn), .B = checkedCalloc(integerPerRow)};
    
#pragma omp parallel for default(shared)
    /* This parallel for might collide with the outernmost parallel for in btc() if the OS doesn't support nested parallel fors */
	for (uint64_t col = 0; col < dim.y; col++) {
		uint64_t* d = checkedCalloc(dim.y);
		uint64_t* b = checkedCalloc(integerPerRow);
		for (uint64_t i = 0; i < integerPerColumn; i++) {
			const uint64_t a = *(A + integerPerColumn*col + i);
			for (uint64_t j = 0; j < dim.y; j++) {
				d[j] += POPCNT64(*(A + integerPerColumn*j + i) ^ a);
			}
		}
		uint64_t simi = dim.y*dim.x;
		for (uint64_t j = 0; j < dim.y; j++) {
			if (d[j] < colsums[j]) {
				b[j/64] |= (ONE << j%64);
				simi -= d[j];
			} else {
				simi -= colsums[j];
			}
		}
		if (simi > bestSim) {
            /* Check for the condition before going into the expensive critical block */
#pragma omp critical
            {
                if (simi > bestSim) {
                    /* We must re-check for simi as other thread might have changed it outside the critical */
                    memcpy(bestVectors.A, &A[integerPerColumn*col], integerPerColumn*sizeof(uint64_t));
                    memcpy(bestVectors.B, b, integerPerRow*sizeof(uint64_t));
                    bestSim = simi;
                }
            }
		}
		free(d);
		free(b);
	}

	return bestVectors;
}

/*
 * Boolean tensor decomposition.
 * Calculates a decomposition of a 3-way tensor by clustering one mode and
 * making a rank-1 decomposition of the other two.
 * The output is composed of three matrices of dimension x*k, y*k and z*k
 * where x,y,z are the dimensions of the input tensor X and k denotes the
 * number of clusters.
 */
decom3 btc_parallel(const bitwiseTensor* X, const dim3 dim, const uint64_t numberOfClusters, const uint64_t reps, const uint64_t numThreads) {
	if (numThreads != 0)
		omp_set_num_threads(numThreads);
	//printf("max threads:%i, numThreads:%lu\n", omp_get_max_threads(), numThreads);
	const uint64_t integerPerColumn = numberOfUint64s(dim.x);
	const uint64_t integerPerRow = numberOfUint64s(dim.y);
	const uint64_t integerPerTube = numberOfUint64s(dim.z);
	decom3 result = {.A=checkedCalloc(integerPerColumn*numberOfClusters),
			.B=checkedCalloc(integerPerRow*numberOfClusters),
			.C=checkedCalloc(integerPerTube*numberOfClusters), .simi=0};

	for (uint64_t rep = 0; rep < reps; rep++){
		// select numberOfClusters x*y slices of X and assign each to its closest centroid
		uint64_t* permutedZ = randomPermutation(dim.z);
#ifdef TESTMODE
		printf("Warning: non random choice of cluster centroids.\n");
		permutedZ = range(dim.z, rep);
#endif
		decom centroids[numberOfClusters];
		uint64_t* d = checkedCalloc(dim.z*numberOfClusters); // distances to centroids
#pragma omp parallel for
		for (uint64_t k = 0; k < numberOfClusters; k++) {
			//printf("threads:%i -- %i\n",omp_get_num_threads(),omp_get_thread_num());
			const uint64_t slice = permutedZ[k];
			centroids[k] = rank1approx(&(X->tensor[slice*dim.y*integerPerColumn]), &(X->onesPerColumn[slice*dim.y]), (dim2) {dim.x, dim.y});
#pragma omp parallel for 
            // This parallel for might collide with the above one if the OS doesn't support nested parallel fors
			for (uint64_t r = 0; r < dim.z; r++) {
				for (uint64_t j = 0; j < dim.y; j++) {
					if (centroids[k].B[j/64] & (ONE << j%64)) {
						for (uint64_t i = 0; i < integerPerColumn; i++) {
							d[k*dim.z+r] += POPCNT64(X->tensor[r*dim.y*integerPerColumn + integerPerColumn*j + i]
							                   ^ centroids[k].A[i]);
						}
					} else {
						d[k*dim.z+r] += X->onesPerColumn[r*dim.y + j];
					}
				}
			}
		}
		free(permutedZ);

		uint64_t simi = 0;
		uint64_t* clusterLabels = checkedCalloc(dim.z);
#pragma omp parallel for reduction(+:simi)
		for (uint64_t r = 0; r < dim.z; r++) {
			uint64_t minD = UINT64_MAX;
			int indexOfMinD = INT_MAX;
			for (uint64_t k = 0; k < numberOfClusters; k++) {
				if (d[k*dim.z+r] < minD) {
					minD = d[k*dim.z+r];
					indexOfMinD = k;
				}
			}
			uint64_t s = dim.x*dim.y - minD;
			simi += s;
			clusterLabels[r] = indexOfMinD;
		}
		free(d);

		if (simi > result.simi) {
			result.simi = simi;
			memset(result.C, 0, integerPerTube*numberOfClusters*sizeof(uint64_t));
			for (uint64_t k = 0; k < numberOfClusters; k++) {
				memcpy(&result.A[integerPerColumn * k], centroids[k].A, integerPerColumn*sizeof(uint64_t));
				memcpy(&result.B[integerPerRow * k], centroids[k].B, integerPerRow*sizeof(uint64_t));
//TODO: The cluster assignment matrix obtained is bogus with the current parallelization!
				fprintf(stderr, "Do not use btc_parallel() until the parallelization is fixed!");
#pragma omp parallel for
				for (uint64_t r = 0; r < dim.z; r++) {
					if (clusterLabels[r] == k) {
						result.C[integerPerTube*k + r/64] |= ONE << r % 64;
					}
				}
			}
		}
		for (uint64_t k = 0; k < numberOfClusters; k++) {
			free(centroids[k].A);
			free(centroids[k].B);
		}
		free(clusterLabels);
	}

	return result;
}
