#include <string.h>
#include "mex.h"
#include "btc.h"
//#include <simstruc.h> // needed to forward printf output to matlab

/* Given a binary 3-way tensor, returns a clustering.
 *
 * This function is intended to be called from Matlab.
 */
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
	if (!(nrhs == 7))
        mexErrMsgTxt("Wrong number of input arguments.");
    if (nlhs != 3)
        mexErrMsgTxt("Wrong number of output arguments. Exactly four arguments expected.");

	const double const *mdim = mxGetData(prhs[4]);
    const dim3 dim = { .x = mdim[0], .y = mdim[1], .z = mdim[2]};
    const mwSize numberOfClusters = mxGetScalar(prhs[5]);
    const mwSize oldSim = mxGetScalar(prhs[6]);

    const size_t m = mxGetM(prhs[0]);
    const size_t mA = mxGetM(prhs[1]);
    const size_t mB = mxGetM(prhs[2]);
    const size_t mC = mxGetM(prhs[3]);

    const double* sT = mxGetData(prhs[0]);
    const double* sA = mxGetData(prhs[1]);
    const double* sB = mxGetData(prhs[2]);
    const double* sC = mxGetData(prhs[3]);

   /* Convert to bitwise representation and run iterative updates. */
    bitwiseTensor* T = sparseToBitRepresentation(sT, m, dim);
    bitwiseTensor* A = sparseToBitRepresentation(sA, mA, (dim3) {dim.x, numberOfClusters, 1});
    bitwiseTensor* B = sparseToBitRepresentation(sB, mB, (dim3) {dim.y, numberOfClusters, 1});
    bitwiseTensor* C = sparseToBitRepresentation(sC, mC, (dim3) {dim.z, numberOfClusters, 1});

    const uint64_t integerPerColumn = numberOfUint64s(dim.x);
	const uint64_t integerPerRow = numberOfUint64s(dim.y);
	const uint64_t integerPerTube = numberOfUint64s(dim.z);

	decom3 btcRes = {
		.A=checkedCalloc(integerPerColumn*numberOfClusters),
		.B=checkedCalloc(integerPerRow*numberOfClusters),
		.C=checkedCalloc(integerPerTube*numberOfClusters),
		.simi=oldSim};
    memcpy(btcRes.A , A->tensor, integerPerColumn*numberOfClusters*sizeof(uint64_t));
    memcpy(btcRes.B , B->tensor, integerPerRow*numberOfClusters*sizeof(uint64_t));
    memcpy(btcRes.C , C->tensor, integerPerTube*numberOfClusters*sizeof(uint64_t));

    free(A->tensor);
    free(A->onesPerColumn);
    free(A);
    free(B->tensor);
    free(B->onesPerColumn);
    free(B);
    free(C->tensor);
    free(C->onesPerColumn);
    free(C);

	clustering c = {
			.centroids = checkedCalloc(integerPerColumn*dim.y*numberOfClusters),
			.assignment = checkedCalloc(integerPerTube*numberOfClusters),
			.similarity = 0};

	for (uint64_t k = 0; k < numberOfClusters; k++) {
		uint64_t* centr = outerProduct(
				&btcRes.A[integerPerColumn * k],
				&btcRes.B[integerPerRow * k],
				(dim2) {dim.x, dim.y});
		memcpy(&c.centroids[integerPerColumn * dim.y * k], centr,
				integerPerColumn*dim.y*sizeof(uint64_t));
		free(centr);
	}
	memcpy(c.assignment, btcRes.C, integerPerTube*numberOfClusters*sizeof(uint64_t));
	c.similarity = btcRes.simi;

    iterativeUpdates(T,&c,dim,numberOfClusters,true);

    free(T->tensor);
    free(T->onesPerColumn);
    free(T);

    /* Prepare output. (Dense!) */
	const mwSize dimCentroids[3] = {dim.x, dim.y, numberOfClusters};
    const mwSize dimAssignment[2] = {dim.z, numberOfClusters};

    mxLogical* centroids = (mxLogical*) toMatlabRepresentation(
    		c.centroids, (dim3) {dim.x,dim.y,numberOfClusters});

    plhs[0] = mxCreateLogicalArray(3, dimCentroids);
    mxLogical *pnA = mxGetLogicals(plhs[0]);
    memcpy(pnA, centroids, dim.x*dim.y*numberOfClusters*sizeof(mxLogical));

    mxLogical* assignment = (mxLogical*) toMatlabRepresentation(
    		c.assignment, (dim3) {dim.z,numberOfClusters,1});

    plhs[1] = mxCreateLogicalArray(2, dimAssignment);
    mxLogical *pnC = mxGetLogicals(plhs[1]);
    memcpy(pnC, assignment, dim.z*numberOfClusters*sizeof(mxLogical));

    plhs[2] = mxCreateDoubleScalar(c.similarity);

    free(btcRes.A);
    free(btcRes.B);
    free(btcRes.C);
    return;
}
