#if __STDC_VERSION__ >= 199901L
#define _XOPEN_SOURCE 600
#else
#define _XOPEN_SOURCE 500
#endif /* __STDC_VERSION__ */
#include <errno.h>
#include <omp.h>
#include <getopt.h>
#include <string.h>
#include <stdio.h>
#include <sys/time.h>
#include <time.h>
#include <stdlib.h>
#include <stdbool.h>
#include <x86intrin.h>
#include "btc.h"

/*
 * Opens file. Prints an error if opening failed.
 */
FILE * checkedFopen(const char* filename, const char* mode) {
	FILE *fp = fopen(filename, mode);
	if (fp == NULL) {
		perror("Failed to open file");
	}

	return fp;
}

void runBTC(const char* dataFile, const char* outputPrefix, const int k, const int reps) {
	FILE* fp = checkedFopen(dataFile, "rb");
	fseek(fp, 0L, SEEK_END);
	long int fileLength = ftell(fp);
	size_t doublesInFile = fileLength/sizeof(double);
	size_t doublesPerRow = doublesInFile / 3 - 1;
	fseek(fp, 0L, SEEK_SET);
	double* sparseTensor = malloc(fileLength);
	fread(sparseTensor, sizeof(double), doublesInFile, fp);
	fclose(fp);

	// sparseTensor holds the coordinates of the ones in three concatenated columns
	// the last row of the input file contains the tensor dimensions
	dim3 dim = {.x=sparseTensor[doublesInFile-3],
			    .y=sparseTensor[doublesInFile-2],
			    .z=sparseTensor[doublesInFile-1]};

	struct timespec tstart,tend;
	double dif;
	bitwiseTensor *bTens = sparseToBitRepresentation(sparseTensor,doublesPerRow, dim);
	free(sparseTensor);
	clock_gettime(CLOCK_MONOTONIC, &tstart);
	decom3 res = btc(bTens, dim, k, reps, 0);
	clock_gettime(CLOCK_MONOTONIC, &tend);
	dif = (tend.tv_sec - tstart.tv_sec);
	dif += (tend.tv_nsec - tstart.tv_nsec) / 1000000000.0;
	free(bTens->tensor);
	free(bTens->onesPerColumn);
	free(bTens);

	arrayWithLength* A,*B,*C;
	A = toSparseMatlabRepresentation(res.A,(dim3) {dim.x,k,1});
	free(res.A);
	B = toSparseMatlabRepresentation(res.B, (dim3) {dim.y,k,1});
	free(res.B);
	C = toSparseMatlabRepresentation(res.C, (dim3) {dim.z,k,1});
	free(res.C);

	// print results
	char filename[200];
	sprintf(filename, "%s_%i.", outputPrefix, k);
	char fn[200];
	FILE *f;

	strcpy(fn, filename);
	strcat(fn, "A");
	f = checkedFopen(fn, "wb");
	// no need to write the third dimension, A is 2-dimensional
	fwrite(A->data, sizeof(double), A->length*2/3, f);
	fclose(f);
	free(A->data);
	free(A);

	strcpy(fn, filename);
	strcat(fn, "B");
	f = checkedFopen(fn, "wb");
	fwrite(B->data, sizeof(double), B->length*2/3, f);
	fclose(f);
	free(B->data);
	free(B);

	strcpy(fn, filename);
	strcat(fn, "C");
	f = checkedFopen(fn, "wb");
	fwrite(C->data, sizeof(double), C->length*2/3, f);
	fclose(f);
	free(C->data);
	free(C);

	strcpy(fn, filename);
	strcat(fn, "sim");
	f = checkedFopen(fn, "wb");
	fwrite(&res.simi, sizeof(int64_t), 1, f);
	fclose(f);

	strcpy(fn, filename);
	strcat(fn, "time");
	f = checkedFopen(fn, "wb");
	fwrite(&dif, sizeof(double), 1, f);
	fclose(f);
}

void printHelp() {
	printf("Usage:\n");
	printf("  -i  --input               input data for BTC\n");
	printf("  -o  --output              name for output of BTC (endings will be added)\n");
	printf("  -r  --reps                number of repetitions\n");
	printf("  -k  --clusters            number of clusters\n");
	printf("  -h, --help                print this help and exit\n");
	printf("\n");
	printf("To run BTC on your own data, input, output, clusters, and\n"
			"repetitions need to be specified.\n\n");
	printf("Your data must be a 3-way binary tensor. The input file should\n"
			"hold the coordinates of the non-zeros as doubles in column-major\n"
			"order. Additionally, The last three doubles in the input file\n"
			"need to specify the dimensionality of the original tensor.\n\n");
	printf("The output will consist of five files, three of which contain\n"
			"the tensor decomposition, file endings .A, .B, and .C . If the\n"
			"tensor dimension is x-y-z, A has size x-k, B has size y-k, and\n"
			"C has size z-k. While C is a cluster assignment matrix and is\n"
			"stored as dense (one bool/int8 per field), A and B are stored\n"
			"as sparse matrices in column-major order. The remaining two\n"
			"files hold one value each: .sim holds the achieved similarity\n"
			"as an int64 value. .time holds the time it took to obtain the\n"
			"decomposition as double.\n\n");
}

/*
 * To use the predefined experiments, this function needs to be called such that the
 * folder "experiments" is on the same level as the folder the code is called from.
 */
int main(int argc, char **argv) {

	int c;
	const char * short_opt = "hi:o:r:k:";
	struct option long_opt[] = {
			{ "help", no_argument, NULL, 'h' },
			{ "input", required_argument, NULL, 'i' },
			{ "output", required_argument, NULL, 'o' },
			{ "reps", required_argument, NULL, 'r' },
			{ "clusters", required_argument, NULL, 'k' },
			{ NULL, 0, NULL, 0 } };

	char input[255] = "";
	char output[255] = "";
	int reps=0;
	int k=0;

	while ((c = getopt_long(argc, argv, short_opt, long_opt, NULL )) != -1) {
		switch (c) {
			case -1: /* no more arguments */
			case 0: /* long options toggles */
				break;

			case 'i':
				strncpy(input,optarg, 255);
				break;

			case 'o':
				strncpy(output,optarg, 255);
				break;

			case 'r':
				reps = atoi(optarg);
				break;

			case 'k':
				k = atoi(optarg);
				break;

			case 'h':
				printHelp();
				return (0);

			case '?':
				fprintf(stderr, "Try `%s --help' for more information.\n", argv[0]);
				return (-2);

			default:
				fprintf(stderr, "%s: invalid option -- %c\n", argv[0], c);
				fprintf(stderr, "Try `%s --help' for more information.\n", argv[0]);
				return (-2);
		};
	}

	if (reps > 0 && k > 0 && strlen(input) > 0 && strlen(output) > 0) {
		printf("BTC started.\n");
		runBTC(input, output, k, reps);
		printf("BTC done.\n");
		fflush(stdout);
	} else {
		fprintf(stderr, "You did not specify all necessary parameters to run BTC on your data.\n\n");
		printHelp();
	}

	return (0);
}

