function [L, LE] = dl_btc(T, varargin)
% DL_BTC - computes description length for BTC
%
% Usage:
% L = DL_BTC(T, A, B, C)
%   Computes the description length of a Boolean tensor clustering of
%   [sp]tensor T using factors A and B and cluster assignment C. A and B
%   must have at least one non-zero entry. C can be either a binary matrix 
%   with exactly one non-zero per row or a an array of cluster indices.
%
% L = DL_BTC(T) is equivalent to DL_BTC(T, [], [], []) and returns the
%   description length of T when no model is used.
%
% L = DL_BTC(T, ..., 'EmptyClusters', true/false) controls whether the
%   method assumes there can be empty clusters or not. Default = true.
%
% [L, LE] = DL_BTC(T, A, B, C)
%   Also returns the error part of the description length.
%
% See also: BCLUST2SPTENS, NCHOOSEKLOG2

%% Process input arguments

p = inputParser;
defaultA = [];
defaultB = [];
defaultC = [];
defaultEmptyClusters = true;

checkT = @(x) all([any([isa(x, 'sptensor'), isa(x, 'tensor')]), ndims(T) == 3]);
checkA = @(x) validateattributes(x, {'numeric'}, {'2d', 'binary'}, 'dl_btc', 'A');
checkB = @(x) validateattributes(x, {'numeric'}, {'2d', 'binary'}, 'dl_btc', 'B');
checkC = @(x) true;
%@(x) any([all([isinteger(x), isvector(x)]), ...
    %all([max(x(:)) == 1, min(x(:)) == 0, all(sum(x, 2) == 1)])]);
checkEmptyClusters = @(x) validateattributes(x, {'logical'}, {'scalar'}, 'dl_btr', 'EmptyClusters');

addRequired(p, 'T', checkT);
addOptional(p, 'A', defaultA, checkA);
addOptional(p, 'B', defaultB, checkB);
addOptional(p, 'C', defaultC, checkC);
addParameter(p, 'EmptyClusters', defaultEmptyClusters, checkEmptyClusters);

parse(p, T, varargin{:});

A = p.Results.A;
B = p.Results.B;
C = p.Results.C;
allowEmptyClusters = p.Results.EmptyClusters;


%% Get parameters and their cost

sz = size(T);
n = sz(1);
m = sz(2);
l = sz(3);

if isempty(C)
    k = 0;
elseif ~isvector(C)
    k = size(C, 2);
    clusters = nan(size(C,1), 1);
    for i=1:k
        I = find(C(:, i));
        clusters(I) = i;
    end
    C = clusters;
else
    if size(C, 2) > size(C, 1)
        C = C';
    end
    k = max(C);
end

% If we have empty clusters, check if we have any, make sure k is correct,  
% and nullify the centroids of the empty clusters
if allowEmptyClusters
    k = size(A, 2);
    if size(B, 2) ~= k
        error('The number of columns in A and B do not match (%i vs %i)', k, size(B, 2));
    end
    presentClusters = unique(C);
    emptyClusters = setdiff(1:k, presentClusters);
    for c=emptyClusters
        % Nullify the empty cluster centroids
        A(:, c) = 0;
        B(:, c) = 0;
    end
end
    

% Checks for dimensions
if k > l
    error('There are more clusters than frontal slices (%i vs %i)', k, l);
end
if size(A, 2) ~= k
    error('The number of clusters does not agree with the number of factors in A (%i vs %i)', k, size(A, 2));
end
if size(B, 2) ~= k
    error('The number of clusters does not agree with the number of factors in B (%i vs %i)', k, size(B, 2));
end
if k > 0
    % These tests make only sense for non-zero factorizations
    if size(A, 1) ~= n
        error('The number of rows in A does not agree with the number of rows in T (%i vs %i)', size(A, 1), n);
    end
    if size(B, 1) ~= m
        error('The number of columns in B does not agree with the number of columns in T (%i vs %i)', size(B, 1), m);
    end
    if size(C, 1) ~= l
        error('The number of slices in C does not agree with the number of slices in T (%i vs %i)', size(C, 1), l);
    end
end

% The L^0 cost of n, m, and l
L_n = L0(n);
L_m = L0(m);
L_l = L0(l);

% There should not be more than l clusters
L_k = log2(l);

%% Factorization costs
if k > 0
    [L_A, L_B] = DtMModel(A, B, n, m, k);
    L_C = l*log2(k);
else
    L_A = 0;
    L_B = 0;
    L_C = 0;
end

% Model part of L
LM = L_n + L_m + L_l + L_k + L_A + L_B + L_C;

%% Error costs
if k > 0
    Ta = bclust2sptens(A, B, C);
    E = xor(T, Ta);
    Tasum = nnz(Ta);
    undercover = nchooseklog2(n*m*l-Tasum, nnz(E > Ta));
    overcover  = nchooseklog2(Tasum, nnz(and(E, Ta)));
else
    Tasum = 0;
    E = T;
    undercover = nchooseklog2(n*m*l, nnz(E));
    overcover  = nchooseklog2(0, 0);
end


if Tasum == 0 || Tasum == n*m*l
    TaSumContrib = log2(n*m*l);
else
    TaSumContrib = log2(Tasum) + log2(n*m*l - Tasum);
end
%Csum = sum(C(:));

LE = undercover + overcover + TaSumContrib;


%% Final sum
L = LM + LE;

return
%% end of the function!

%% L^0 cost function
function l = L0(n)
l = 0;
if n == 0
    return;
end
tmp = log2(n);
while tmp > 0
    l = l + tmp;
    tmp = log2(tmp);
end
l = l + log2(2.865064);


%% Model cost function
function [L_A, L_B] = DtMModel(A, B, n, m, k)
if k == 0
    L_A = 0;
    L_B = 0;
    return;
end
L_A = k*log2(n) + sum(nchooseklog2(n, sum(A)));
L_B = k*log2(m) + sum(nchooseklog2(m, sum(B,2)'));

